declare namespace saxon="http://saxon.sf.net/";
declare namespace java="http://saxon.sf.net/java-type";

(: declare document :)
declare variable $doc as xs:string external;
declare variable $document := doc($doc)/osm; 
declare variable $indexedNodes
   as java:net.sf.saxon.extra.IndexedSequence
   := saxon:index($document/node, saxon:expression("@id"));

declare function local:dotproduct($v1, $v2, $w1, $w2)
{
  (: function to compute the dotproduct, it is multiplied by 10^7
     to get rid of rounding errors :)
  let $dxv := $v2/@lat - $v1/@lat
  let $dyv := $v2/@lon - $v1/@lon
  let $dxw := $w2/@lat - $w1/@lat
  let $dyw := $w2/@lon - $w1/@lon
  return (10000000*(($dxv*$dxw) + ($dyv*$dyw)))
};

declare function local:distance($n1, $n2) 
{
  (: computes the square of the L2 distance between two points,
     it is multiplied by 10^10 to get rid of rounding errors :)
  let $a := (data($n1/@lon) - data($n2/@lon))
  let $b := (data($n1/@lat) - data($n2/@lat))
  return round(10000000000*(($a*$a)+($b*$b)))
};

declare function local:minnode 
  ( $nodes as node()* )  as node()* {
  (: find the node with the minimum distance value :)
   $nodes[. = min($nodes/dist)]
 } ;

declare function local:findfriend($way, $posw, $candidates)
{
  (: assign begin and end node of way $way :)
  let $begin := $way/nd[1]
  let $end := $way/nd[last()]
  (: assign name, ref and type, these are needed for later checks :)
  let $name := $way/tag[@k="name"]/@v
  let $ref := $way/tag[@k="ref"]/@v
  let $type := $way/tag[@k="highway"]/@v
  (: determine possible friends -ways with the same name-, or if no name exists
     ways with the same reference -for example in case of motorways- :)
  let $pfriends :=
    if ($name)
      then 
        (: find possible friends with the same name :)
        $candidates[tag/@k="name"][tag/@v=$name][tag/@k="highway"]
          [tag/@v=$type]
      else 
      (
        if ($ref)
          then 
          (: find possible friends with the same ref :)
            $candidates[tag/@k="ref"][tag/@v=$ref][tag/@k="highway"]
              [tag/@v=$type]
          else ()
      )

  (: type of ways has to be the same :)
  let $trainees := (
    for $tr at $pos in ($pfriends except $way)
      let $begintr := $tr/nd[1]
      let $endtr := $tr/nd[last()]
        (: dist1 is the distance of the first node of a way and the last node 
           of a way with the same name, hopefully -but not necessarily- in 
           the other direction :)
      return (
        let $v1 := saxon:find($indexedNodes, $begin/@ref)
        let $v2 := saxon:find($indexedNodes, $end/@ref)
        let $w1 := saxon:find($indexedNodes, $begintr/@ref)
        let $w2 := saxon:find($indexedNodes, $endtr/@ref)
        (: compute dotproduct to see if the ways are more or less in opposite 
           direction :)
        let $semipar := local:dotproduct($v1, $v2, $w1, $w2)
        return (
          if ($semipar > 0)
            then () (: pruning step, see inside view :)
            else (
              (: distance is the distance between the begin node of the first 
                 way and the end node of the second way, plus the distance 
                 between the end node of the first way and the begin node of 
                 the second way :)
              let $dist1 := local:distance($v1, $w2)
              let $dist2 := local:distance($w1, $v2)
              let $dist := $dist1+$dist2
              return (  
                (: uid is a unique identifier, needed in constructing
                   the relation :)
                if ($dist < 8000)
                then 
                  (: construct intermediate XML node :)
                  <couple uid="{concat('-', $pos, $posw)}">
                    <w1> {$way/@id} </w1>
                    <w2> {$tr/@id} </w2>
                    <dist> { $dist } </dist>
                    <dp sol="{$semipar}"/>
                  </couple>
                else ()
              )
          )
        )
      )
    )
  (: extract the node with the minimum distance, return that as $way's friend :)
  return local:minnode($trainees)
};

declare function local:coupling($input, $output)
{
  (: return pair <x,y> iff <x,y> in $input and <y,x> in $input :)
  (: this is done in order to assure relation in bijective :)
  if (count($input) = 0) 
    then $output
    else
      let $first := $input[1]
      let $current := $first/w1/@id
      let $friend := $first/w2/@id
      (: -current, friend- pair is in input, check if -friend, current- also
         exists in input :)
      let $corresponding := $input[(w2/@id = $current) and (w1/@id=$friend)]
      return (
        (: if matching pair is found, remove it from the input, and add pair
           to the output, else do not add it to the output :)
        if ($corresponding)
          then local:coupling((($input except $first) except $corresponding),
                              $output union $first)
          else local:coupling(($input except $first),$output) 
      )
};

declare function local:parseresults($results)
{
  (: extra function needed in order to generate results
     this is done so the whole algorithm can use a richer way-node :)
  (: input  : results in internal format
     output : relations as described by the osm proposal :)
  for $way in $results
    return (
      <relation action="modify" visible="true" id="{$way/@uid}">
       <member type="way" ref="{$way/w1/@id}" role=""/> 
       <member type="way" ref="{$way/w2/@id}" role=""/> 
       <tag k="type" v="dual_carriageway"/>
      </relation>
    )
};
<osm version='0.5' generator='JOSM'>
{
  let $candidates := ( 
    for $y in $document/way
      (: filter relevant ways, that are primary/secondary/tertiary ways, 
         which are oneway (oneway=1/yes/true) or motorways and not roundabouts,
         because roundabouts can give trouble since they are circular :)
      let $oneway := $y/tag[@k="oneway"]/@v
      let $valuetag := $y/tag/@v
      where ((($oneway = "1") or ($oneway = "yes") or ($oneway = "true")) and 
              ($y/tag/@k = "highway") and (($valuetag = "primary") or 
              ($valuetag = "secondary") or ($valuetag = "tertiary") or
              ($valuetag = "motorway") or ($valuetag = "trunk")) 
              and not(($y/tag/@k = "junction") and ($y/tag/@v = "roundabout")))
        return $y 
  )
  (: construct a list of pairs :)
  let $list := (
  for $y at $pos in $candidates
    return local:findfriend($y, $pos, $candidates)
  )
  (: parse the list in order to get useable XML code :)
  (: return document and resulting relation :)
  return $document/* union local:parseresults(local:coupling($list, ()))
}
</osm>
