declare namespace saxon="http://saxon.sf.net/";
declare namespace java="http://saxon.sf.net/java-type";

(: declare document :)
declare variable $doc as xs:string external;
declare variable $document := doc($doc)/osm; 
declare variable $indexedNodes
   as java:net.sf.saxon.extra.IndexedSequence
   := saxon:index($document/node, saxon:expression("@id"));
declare variable $indexedWays
   as java:net.sf.saxon.extra.IndexedSequence
   := saxon:index($document/way, saxon:expression("@id"));

declare function local:transitive($fullway, $jname, $list, $depth, $way)
{
  (: last element $way is needed for lat and lon, which are not in $fullway :)

  (: check if way is a highway :)
  let $c1 := $fullway/tag/@k = "highway"
  (: check if way is a motorway_link :)
  let $c2 := $fullway/tag/@v = "motorway_link"
  (: or trunk_link :)
  let $c2t := $fullway/tag/@v = "trunk_link"
  (: the first time you do not want to leave the recursion since the first way
     always satisfies all conditions :)
  let $c3 := $depth gt 0
  (: check if node corresponding to way has a name, if so recursion can stop :)
  let $last := $fullway/nd[last()]
  let $fnode := saxon:find($indexedNodes, $last/@ref)
  let $c4 := $fnode/tag[@k = "name"]/@v = $jname 
  return (
    (: check all conditions :)
    if ($c1 and ($c2 or $c2t) and $c3 and $c4) 
    (: end recursion if all conditions are satisfied - succesfull trace :)
    then $list 
    else ( 
      if ($depth = 12)
        (: end recursion if a certain depth is reached - unsuccesfull trace :)
        then ()
        else (
          let $ew := $last/@ref
          (: call recursive function for each way of which the first node is 
             equal to the last node of the current way - $fullway - :)
          for $x in $document/way[nd[1]/@ref = $ew]
            let $dist := local:distance($way, local:transform($x))
            where ($dist < 0.0004)
            return local:transitive($x, $jname, $list union $fullway, $depth+1,
                                    $way)
          )
        )
      )
}; 

declare function local:extractexitnumber($nodes)
{
  (: input  : a list of node references :)
  (: output : if there are no two nodes with different exit numbers and at 
              least one node has a reference number, then return that exit 
              number else return empty :)

  (: a tag key ref is stored in -some- nodes, and represents the exit number :)
  let $all := ( (:  all -possibly different- exit codes of 1 junction :)
    for $node in ($nodes/@ref)
      let $nd := saxon:find($indexedNodes, $node)/tag[@k="ref"]/@v
      return ( $nd )
  )
  
  (: check if all values are the same, if not return empty :) 
  return(
    if (count(distinct-values($all))=1)
      then $all[1]
      else ()
    )
};

declare function local:distance($n1, $n2) 
{
  (: computes the square of the L2 distance between two points :)
  let $a := data($n1/lon) - data($n2/lon)
  let $b := data($n1/lat) - data($n2/lat)
  return (($a*$a)+($b*$b))
};

declare function local:transform($y)
{
(: input: a way from the document 
   output: an adapted way, with an average lat and lon :)
    (: first node reference of way $y :)
    let $first := $y/nd[1]
    (: last node reference of way $y :)
    let $last := $y/nd[last()]
    (: lookup first node in document based on reference $first :)
    let $fnode := saxon:find($indexedNodes, $first/@ref) (: unique :)
    (: lookup last node in document based on reference $last :)
    let $lnode := saxon:find($indexedNodes, $last/@ref) (: unique :)
    let $latf := (data($fnode/@lat))
    let $latl := (data($lnode/@lat)) 
    (: longtitude first and last node :)
    let $lonf := (data($fnode/@lon)) 
    let $lonl := (data($lnode/@lon))
    return ( 
          <way> 
             (: compute average latitute and longtitude, which is computed as
                 the average location of the first and last node :) 
            <lat> { ($latf+$latl)*(0.5) } </lat>
            <lon> { ($lonf+$lonl)*(0.5) } </lon>
          </way>
   )
};

declare function local:split($firstway, $junction, $set1, $set2)
{
   (: splits a junction in multiple junctions if the distance between 
      "junction-islands" is more than 0.0004 :)
    if (count($junction) > 0)
      then 
        let $way := $junction[1]
        let $dist := local:distance($way, $firstway) 
        return (
          if ($dist < 0.0004)
            then local:split($firstway, ($junction except $way), 
                             ($set1 union $way), $set2)
            else local:split($firstway, ($junction except $way), 
                             $set1, ($set2 union $way))
         
        )
      else (: no elements left to process :)
        <node> { ($set1 union $firstway) }  </node> union (
        if (count($set2) = 0)
          then ()
          else 
            let $fw2 := $set2[1]
            (: try to split $set2 again, possibly more islands can be 
               discovered :)
            return local:split($fw2, ($set2 except $fw2), (), ())
        )
};

declare function local:makeway($y)
{
  (: input : a way from the document, which is a motorway_link
     output: a self constructed way, which is only created if the first or 
             last node of the given way has a name, and if they both
             have a name, the same name :)
  (: first node reference of way $y :)
  let $first := $y/nd[1]
  (: last node reference of way $y :)
  let $last := $y/nd[last()]
  (: lookup first node in document based on reference $first :)
  let $fnode := saxon:find($indexedNodes, $first/@ref) (: unique :)
  (: lookup last node in document based on reference $last :)
  let $lnode := saxon:find($indexedNodes, $last/@ref) (: unique :)
  (: name associated with first node - possible name of junction :)
  let $fname := $fnode/tag[@k = "name"]/@v
  (: name associated with last node - possible name of junction :)
  let $lname := $lnode/tag[@k = "name"]/@v
  (: latitude first and last node :)
  let $latf := (data($fnode/@lat))
  let $latl := (data($lnode/@lat)) 
  (: longtitude first and last node :)
  let $lonf := (data($fnode/@lon)) 
  let $lonl := (data($lnode/@lon))
  return ( 
    (: at least one of the two nodes should have a name, 
       if both are present they should be the same :)
    if (not($fname and $lname and ($fname != $lname)) and ($fname or $lname))
      then
        (: add way information to result :)
        <way> 
          { $y/@id }
          <tag k="name"> 
            { if ($fname)
                then $fname
                else $lname 
            }
          </tag>
          <nd> { $first/@ref } </nd>
          <nd> { $last/@ref } </nd>
          (: compute average latitute and longtitude, which is computed as
               the average location of the first and last node :) 
          <lat> { ($latf+$latl)*(0.5) } </lat>
          <lon> { ($lonf+$lonl)*(0.5) } </lon>
        </way>
      else ()
  )
};

declare function local:groupjunctions($results)
{
  (: input : a number of ways, which have the format as is used in function
             makeway
     output: groups of junctions, which are grouped by name :)
  for $way in $results
    (: get name of the junction :)
    let $name := $way/tag/@v
    (: find first occurence of a way with name $name :) 
    let $fpos := ($results/tag[@v = $name])[1]
    return (  
      (: the following condition implies that each junction is only 
         treated once :)
      if ($way/tag is $fpos) 
        then 
          <node>
            <location> { $way/tag/@v } </location>
             { (: find ways with the same name as the current way :)
               for $z in $results[tag/@v = $name] 
                 return 
                   <way> { $z/@id } 
                     <begin> { ($z/nd/@ref)[1] } </begin>
                     <end> { ($z/nd/@ref)[2] } </end> 
                     <lat> { data($z/lat) } </lat>
                     <lon> { data($z/lon) } </lon>
                   </way>
             } 
          </node>
        else ()
    )
};

declare function local:makeislands($grouped)
{
  (: input : a list of groups, grouped by name
     output: a list of groups, grouped by name, where all ways in a group 
             are within a distance of 0.0004 :)
  for $junction in $grouped
    (: determine first way in junction, will be used as a reference for the 
       distance computation in the split function :)
    let $firstway := $junction/way[1]
    return ( 
      let $result := local:split($firstway, (($junction/way) except $firstway),
                                 (), ())
      return (
        (: parse result, construct correct tags :)
        for $island in $result
           return ( 
            <node>
            {
            <location> 
              { $junction/location/@v }
            </location> union
             $island/way
            }
            </node>
           )
      )
  )
};

declare function local:distinctways($relation)
{
  for $way in distinct-values($relation)
    (: add way to relation :)
    return <member type='way' ref='{ $way }' role='' /> 
};

declare function local:distinctnodes($relation)
{
  for $node in distinct-values($relation)
    (: add node to relation :)
    return <member type='node' ref='{ $node }' role='' /> 
};

declare function local:constructrelation($way, $name)
{
  (: input : way with a junction name
     output: the transitive closure of that way:)

  (: compute transitive closure, starting from $way :)
  let $cway := $document/way[@id=$way/@id]  
  (: saxon:find($indexedWays, $way/@id) :)
  let $result := local:transitive($cway, $name, (), 0, $way)  
  (: parse results from transitive closure :)
  let $relresult := ( 
    for $wayr in $result
      return (
      <way> { data($wayr/@id) } </way> union 
      <node> { data($wayr/nd[1]/@ref) } </node> union
      <node> { data($wayr/nd[last()]/@ref) } </node> 
    )
  )
  return (
    (: $relresult contains at this point all points from the 
       transitive closure, but the way - $way - itself is not added 
       yet. this is done below :)
    $relresult union
    <way>{ data($way/@id) }</way> union
    <node> { data($way/begin/@ref) } </node> union
    <node> { data($way/end/@ref) } </node>
  ) 
};

<osm version='0.5' generator='JOSM'>
{
(: $results will contain all nodes which will be used for the tagging of 
   motorway junctions and exits, i.e. the nodes that satisfy the condition 
   that at least one of the two nodes should have a name, and if both are 
   present they should be the same :) 
let $results :=
for $y in $document/way
  where (($y/tag/@k = "highway") and (($y/tag/@v = "motorway_link") or 
         ($y/tag/@v = "trunk_link")))
  return local:makeway($y)

(: ways in $result will be grouped according to their junction name by
   function groupjunctions and thereafter into multiple "junction-islands" 
   if needed :)
let $splitgroups := local:makeislands(local:groupjunctions($results)) 

let $relation:= (
  for $junction at $pos in $splitgroups
    (: name of the junction :)
    let $jname := $junction/location/@v
    return ( 
      (: relation tag, with a unique identifier -pos. this is done so that a 
         unique identifier is added later when adding it to the osm database :)
      <relation id='-{$pos}' action='modify' visible='true'> 
      { 
      (: in $onerel all members from one relation are stored :)
      let $onerel := (
      <rawresults> {
      for $way in $junction/way
        return local:constructrelation($way, $jname)
      } </rawresults>
      )
      (: determine unique ways because local:transitive may give duplicates :)
      let $uniqueways := local:distinctways($onerel/way)
      (: determine unique nodes because local:transitive may give duplicates :)
      let $uniquenodes := local:distinctnodes($onerel/node)
      (: determine exit number - if there exist one :)
      let $exit := local:extractexitnumber($uniquenodes)
      return $uniqueways union $uniquenodes union
      ( if ($exit) 
          (: add exit number if there is one :)
          then <tag k='ref' v='{ $exit }'/> 
          else () 
      ) union
         (: add junction tag and name :)
         <tag k='type' v='junction'/> union 
         <tag k='name' v='{$jname}'/>
      }
      </relation>
    )
  )
(: return document and resulting relation :)
return $document/* union $relation 
}

</osm>
